==> capibara_ent-1.1.15-py3-none-any.whl: layers/synthetic_embedding.py <==
"""
Módulo que implementa una capa de Embedding Sintético Mejorada para redes neuronales usando JAX/Flax.

Esta capa combina Capibara SSM, atención multi-cabeza, capas BitNet y Liquid,
y una capa Capibara dispersa para el procesamiento eficiente de las entradas.

Clases:
    EnhancedSyntheticEmbeddingLayer: Implementa una capa de Embedding Sintético Mejorada.

Dependencias:
    - jax: Para operaciones de arrays y diferenciación automática.
    - flax: Para definiciones de módulos de redes neuronales.
"""

import jax
import jax.numpy as jnp
from flax import linen as nn
from typing import Optional
import logging

from .bitnet import BitNet
from .liquid import Liquid
from .sparse_capibara import SparseCapibara
from .bitnet_quantizer import BitNetQuantizer
from .meta_bamdp import MetaBAMDP
from .capibara_jax_ssm import CapibaraJAXSSM
from .capibara2 import Capibara2

logger = logging.getLogger(__name__)


class SyntheticEmbedding(nn.Module):
    """
    EnhancedSyntheticEmbeddingLayer: Una capa optimizada para embeddings sintéticos,
    diseñada para ejecución eficiente en TPUs.

    Esta capa implementa una secuencia de operaciones incluyendo Capibara SSM, atención multi-cabeza,
    capas BitNet y Liquid, Capibara dispersa y cuantización.

    Atributos:
        dim (int): Dimensión de entrada y salida.
        dropout_rate (float): Tasa de dropout para regularización.
        use_residual (bool): Si se usa una conexión residual.
        num_heads (int): Número de cabezas de atención.
        capibara_dim (int): Dimensión de Capibara SSM.
        num_bitnet_layers (int): Número de capas BitNet.
        num_liquid_layers (int): Número de capas Liquid.
    """
    dim: int
    dropout_rate: float = 0.1
    use_residual: bool = True
    num_heads: int = 4
    capibara_dim: int = 16
    num_bitnet_layers: int = 3
    num_liquid_layers: int = 3

    def setup(self):
        self.capibara_jax_ssm = SparseCapibara(dim=self.dim)
        self.attention = nn.SelfAttention(
            num_heads=self.num_heads,
            dropout_rate=self.dropout_rate
        )
        self.bitnet_layers = [BitNet(in_dim=self.dim, out_dim=self.dim)
                              for _ in range(self.num_bitnet_layers)]
        self.liquid_layers = [Liquid(dim=self.dim)
                              for _ in range(self.num_liquid_layers)]
        self.sparse_capibara = SparseCapibara(dim=self.dim)
        self.capibara2 = Capibara2(dim=self.dim)
        self.bitnet_quantizer = BitNetQuantizer()
        self.meta_bamdp = MetaBAMDP()
        self.final_fc = nn.Dense(features=self.dim)

    @nn.compact
    def __call__(self, x: jnp.ndarray, mask: jnp.ndarray = None, training: bool = True) -> jnp.ndarray:
        """
        Forward pass of the EnhancedSyntheticEmbeddingLayer.

        Args:
            x (jnp.ndarray): Input array of shape (batch_size, seq_len, dim).
            mask (jnp.ndarray, optional): Mask array of shape (batch_size, seq_len).
            training (bool): Whether the model is in training mode.

        Returns:
            jnp.ndarray: Output array of shape (batch_size, seq_len, dim).
        """
        self._validate_input(x)
        logger.debug(f"Input shape: {x.shape}, dtype: {x.dtype}")

        if mask is not None:
            self._validate_mask(mask, x.shape)
            x = x * mask[:, :, jnp.newaxis]

        residual = x if self.use_residual else None

        # Apply Capibara SSM
        x = self.capibara_ssm(x)

        # Apply multi-head attention
        x = self.attention(x, x, x, mask=mask, training=training)

        # Apply BitNet and Liquid layers
        def apply_bitnet_liquid(bitnet, liquid, x):
            x = jnp.transpose(bitnet(jnp.transpose(x, (0, 2, 1))), (0, 2, 1))
            return liquid(x)

        x = jax.vmap(apply_bitnet_liquid, in_axes=(0, 0, None))(
            jnp.array(self.bitnet_layers),
            jnp.array(self.liquid_layers),
            x
        )

        # Apply sparse Capibara
        x = self.sparse_capibara(x)

        # Apply second Capibara SSM
        x = self.capibara_ssm2(x)

        # Apply BitNet quantizer
        x = self.bitnet_quantizer(x)

        # Apply Meta BAMDP
        x = self.meta_bamdp(x)

        # Final fully connected layer
        x = self.final_fc(x)

        if self.use_residual:
            x = x + residual

        x = nn.LayerNorm()(x)
        x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=not training)

        logger.debug(f"Output shape: {x.shape}")
        return x

    def _multi_head_attention(self, x, mask, training):
        """Simplified implementation of multi-head attention."""
        batch_size, seq_len, _ = x.shape
        head_dim = self.dim // self.num_heads

        def attention_head(q, k, v):
            attn_weights = jnp.matmul(
                q, k.transpose(-1, -2)) / jnp.sqrt(head_dim)
            if mask is not None:
                attn_weights = jnp.where(
                    mask[:, None, None, :], attn_weights, -1e9)
            attn_weights = jax.nn.softmax(attn_weights, axis=-1)
            return jnp.matmul(attn_weights, v)

        q = nn.Dense(features=self.dim, use_bias=False)(x)
        k = nn.Dense(features=self.dim, use_bias=False)(x)
        v = nn.Dense(features=self.dim, use_bias=False)(x)

        q = q.reshape(batch_size, seq_len, self.num_heads,
                      head_dim).transpose(0, 2, 1, 3)
        k = k.reshape(batch_size, seq_len, self.num_heads,
                      head_dim).transpose(0, 2, 1, 3)
        v = v.reshape(batch_size, seq_len, self.num_heads,
                      head_dim).transpose(0, 2, 1, 3)

        heads = jax.vmap(attention_head)(q, k, v)
        heads = heads.transpose(0, 2, 1, 3).reshape(
            batch_size, seq_len, self.dim)

        return nn.Dense(features=self.dim)(heads)

    def _validate_input(self, x: jnp.ndarray):
        """Validate the input array dimensions."""
        if x.ndim != 3:
            error_msg = f"Expected input array with 3 dimensions (batch_size, seq_len, dim), but got {
                x.ndim} dimensions."
            logger.error(error_msg)
            raise ValueError(error_msg)
        if x.shape[-1] != self.dim:
            error_msg = f"Expected dim to be {
                self.dim}, but got {x.shape[-1]}."
            logger.error(error_msg)
            raise ValueError(error_msg)

    def _validate_mask(self, mask: jnp.ndarray, input_shape: tuple):
        """Validate the mask array dimensions."""
        if mask.ndim != 2:
            error_msg = f"Expected mask array with 2 dimensions (batch_size, seq_len), but got {
                mask.ndim} dimensions."
            logger.error(error_msg)
            raise ValueError(error_msg)
        if mask.shape != input_shape[:-1]:
            error_msg = f"Expected mask shape to be {
                input_shape[:-1]}, but got {mask.shape}."
            logger.error(error_msg)
            raise ValueError(error_msg)

    def get_config(self) -> dict:
        """
        Get the configuration of the EnhancedSyntheticEmbeddingLayer.

        Returns:
            dict: A dictionary containing the layer's configuration.
        """
        return {
            "dim": self.dim,
            "dropout_rate": self.dropout_rate,
            "use_residual": self.use_residual,
            "num_heads": self.num_heads,
            "capibara_dim": self.capibara_dim,
            "num_bitnet_layers": self.num_bitnet_layers,
            "num_liquid_layers": self.num_liquid_layers
        }


# Example usage
if __name__ == "__main__":
    try:
        # Set up logging
        logging.basicConfig(level=logging.DEBUG)

        # Create a sample input array
        batch_size, seq_len, dim = 32, 10, 256
        key = jax.random.PRNGKey(0)
        x = jax.random.normal(key, (batch_size, seq_len, dim))

        # Create a sample mask
        mask = jnp.ones((batch_size, seq_len))
        mask = mask.at[:, 5:].set(0)  # Mask out the last 5 tokens

        # Initialize the EnhancedSyntheticEmbeddingLayer
        layer = SyntheticEmbedding(dim=256)

        # Initialize parameters
        params_key, dropout_key = jax.random.split(key)
        params = layer.init(params_key, x, mask, training=True)

        # Perform forward pass
        output = layer.apply(params, x, mask, training=True,
                             rngs={'dropout': dropout_key})

        print(f"Input shape: {x.shape}")
        print(f"Output shape: {output.shape}")
        print(f"Layer config: {layer.get_config()}")

        logger.info(
            "EnhancedSyntheticEmbeddingLayer example completed successfully")
    except Exception as e:
        logger.exception(
            f"An error occurred during the EnhancedSyntheticEmbeddingLayer example: {str(e)}")

==> capibara_ent-1.2.1-py3-none-any.whl: layers/synthetic_embedding.py <==
"""
Módulo que implementa una capa de Embedding Sintético Mejorada para redes neuronales usando JAX/Flax.

Esta capa combina Capibara SSM, atención multi-cabeza, capas BitNet y Liquid,
y una capa Capibara dispersa para el procesamiento eficiente de las entradas.

Clases:
    EnhancedSyntheticEmbeddingLayer: Implementa una capa de Embedding Sintético Mejorada.

Dependencias:
    - jax: Para operaciones de arrays y diferenciación automática.
    - flax: Para definiciones de módulos de redes neuronales.
"""

import jax  # type: ignore
import jax.numpy as jnp  # type: ignore
from flax import linen as nn  # type: ignore
from typing import Optional
import logging

from .bitnet import BitNet  # type: ignore
from .liquid import Liquid  # type: ignore
from .sparse_capibara import SparseCapibara  # type: ignore
from .bitnet_quantizer import BitNetQuantizer  # type: ignore
from .meta_bamdp import MetaBAMDP  # type: ignore
from .capibara_jax_ssm import CapibaraJAXSSM  # type: ignore
from .capibara2 import Capibara2  # type: ignore

logger = logging.getLogger(__name__)


class SyntheticEmbedding(nn.Module):
    """
    EnhancedSyntheticEmbeddingLayer: Una capa optimizada para embeddings sintéticos,
    diseñada para ejecución eficiente en TPUs.

    Esta capa implementa una secuencia de operaciones incluyendo Capibara SSM, atención multi-cabeza,
    capas BitNet y Liquid, Capibara dispersa y cuantización.

    Atributos:
        dim (int): Dimensión de entrada y salida.
        dropout_rate (float): Tasa de dropout para regularización.
        use_residual (bool): Si se usa una conexión residual.
        num_heads (int): Número de cabezas de atención.
        capibara_dim (int): Dimensión de Capibara SSM.
        num_bitnet_layers (int): Número de capas BitNet.
        num_liquid_layers (int): Número de capas Liquid.
    """
    dim: int
    dropout_rate: float = 0.1
    use_residual: bool = True
    num_heads: int = 4
    capibara_dim: int = 16
    num_bitnet_layers: int = 3
    num_liquid_layers: int = 3

    def setup(self):
        self.capibara_jax_ssm = SparseCapibara(dim=self.dim)
        self.attention = nn.SelfAttention(
            num_heads=self.num_heads,
            dropout_rate=self.dropout_rate
        )
        self.bitnet_layers = [BitNet(in_dim=self.dim, out_dim=self.dim)
                              for _ in range(self.num_bitnet_layers)]
        self.liquid_layers = [Liquid(dim=self.dim)
                              for _ in range(self.num_liquid_layers)]
        self.sparse_capibara = SparseCapibara(dim=self.dim)
        self.capibara2 = Capibara2(dim=self.dim)
        self.bitnet_quantizer = BitNetQuantizer()
        self.meta_bamdp = MetaBAMDP()
        self.final_fc = nn.Dense(features=self.dim)

    @nn.compact
    def __call__(self, x: jnp.ndarray, mask: jnp.ndarray = None, training: bool = True) -> jnp.ndarray:
        """
        Forward pass of the EnhancedSyntheticEmbeddingLayer.

        Args:
            x (jnp.ndarray): Input array of shape (batch_size, seq_len, dim).
            mask (jnp.ndarray, optional): Mask array of shape (batch_size, seq_len).
            training (bool): Whether the model is in training mode.

        Returns:
            jnp.ndarray: Output array of shape (batch_size, seq_len, dim).
        """
        self._validate_input(x)
        logger.debug(f"Input shape: {x.shape}, dtype: {x.dtype}")

        if mask is not None:
            self._validate_mask(mask, x.shape)
            x = x * mask[:, :, jnp.newaxis]

        residual = x if self.use_residual else None

        # Apply Capibara SSM
        x = self.capibara_ssm(x)

        # Apply multi-head attention
        x = self.attention(x, x, x, mask=mask, training=training)

        # Apply BitNet and Liquid layers
        def apply_bitnet_liquid(bitnet, liquid, x):
            x = jnp.transpose(bitnet(jnp.transpose(x, (0, 2, 1))), (0, 2, 1))
            return liquid(x)

        x = jax.vmap(apply_bitnet_liquid, in_axes=(0, 0, None))(
            jnp.array(self.bitnet_layers),
            jnp.array(self.liquid_layers),
            x
        )

        # Apply sparse Capibara
        x = self.sparse_capibara(x)

        # Apply second Capibara SSM
        x = self.capibara_ssm2(x)

        # Apply BitNet quantizer
        x = self.bitnet_quantizer(x)

        # Apply Meta BAMDP
        x = self.meta_bamdp(x)

        # Final fully connected layer
        x = self.final_fc(x)

        if self.use_residual:
            x = x + residual

        x = nn.LayerNorm()(x)
        x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=not training)

        logger.debug(f"Output shape: {x.shape}")
        return x

    def _multi_head_attention(self, x, mask, training):
        """Simplified implementation of multi-head attention."""
        batch_size, seq_len, _ = x.shape
        head_dim = self.dim // self.num_heads

        def attention_head(q, k, v):
            attn_weights = jnp.matmul(
                q, k.transpose(-1, -2)) / jnp.sqrt(head_dim)
            if mask is not None:
                attn_weights = jnp.where(
                    mask[:, None, None, :], attn_weights, -1e9)
            attn_weights = jax.nn.softmax(attn_weights, axis=-1)
            return jnp.matmul(attn_weights, v)

        q = nn.Dense(features=self.dim, use_bias=False)(x)
        k = nn.Dense(features=self.dim, use_bias=False)(x)
        v = nn.Dense(features=self.dim, use_bias=False)(x)

        q = q.reshape(batch_size, seq_len, self.num_heads,
                      head_dim).transpose(0, 2, 1, 3)
        k = k.reshape(batch_size, seq_len, self.num_heads,
                      head_dim).transpose(0, 2, 1, 3)
        v = v.reshape(batch_size, seq_len, self.num_heads,
                      head_dim).transpose(0, 2, 1, 3)

        heads = jax.vmap(attention_head)(q, k, v)
        heads = heads.transpose(0, 2, 1, 3).reshape(
            batch_size, seq_len, self.dim)

        return nn.Dense(features=self.dim)(heads)

    def _validate_input(self, x: jnp.ndarray):
        """Validate the input array dimensions."""
        if x.ndim != 3:
            error_msg = f"Expected input array with 3 dimensions (batch_size, seq_len, dim), but got {
                x.ndim} dimensions."
            logger.error(error_msg)
            raise ValueError(error_msg)
        if x.shape[-1] != self.dim:
            error_msg = f"Expected dim to be {
                self.dim}, but got {x.shape[-1]}."
            logger.error(error_msg)
            raise ValueError(error_msg)

    def _validate_mask(self, mask: jnp.ndarray, input_shape: tuple):
        """Validate the mask array dimensions."""
        if mask.ndim != 2:
            error_msg = f"Expected mask array with 2 dimensions (batch_size, seq_len), but got {
                mask.ndim} dimensions."
            logger.error(error_msg)
            raise ValueError(error_msg)
        if mask.shape != input_shape[:-1]:
            error_msg = f"Expected mask shape to be {
                input_shape[:-1]}, but got {mask.shape}."
            logger.error(error_msg)
            raise ValueError(error_msg)

    def get_config(self) -> dict:
        """
        Get the configuration of the EnhancedSyntheticEmbeddingLayer.

        Returns:
            dict: A dictionary containing the layer's configuration.
        """
        return {
            "dim": self.dim,
            "dropout_rate": self.dropout_rate,
            "use_residual": self.use_residual,
            "num_heads": self.num_heads,
            "capibara_dim": self.capibara_dim,
            "num_bitnet_layers": self.num_bitnet_layers,
            "num_liquid_layers": self.num_liquid_layers
        }


# Example usage
if __name__ == "__main__":
    try:
        # Set up logging
        logging.basicConfig(level=logging.DEBUG)

        # Create a sample input array
        batch_size, seq_len, dim = 32, 10, 256
        key = jax.random.PRNGKey(0)
        x = jax.random.normal(key, (batch_size, seq_len, dim))

        # Create a sample mask
        mask = jnp.ones((batch_size, seq_len))
        mask = mask.at[:, 5:].set(0)  # Mask out the last 5 tokens

        # Initialize the EnhancedSyntheticEmbeddingLayer
        layer = SyntheticEmbedding(dim=256)

        # Initialize parameters
        params_key, dropout_key = jax.random.split(key)
        params = layer.init(params_key, x, mask, training=True)

        # Perform forward pass
        output = layer.apply(params, x, mask, training=True,
                             rngs={'dropout': dropout_key})

        print(f"Input shape: {x.shape}")
        print(f"Output shape: {output.shape}")
        print(f"Layer config: {layer.get_config()}")

        logger.info(
            "EnhancedSyntheticEmbeddingLayer example completed successfully")
    except Exception as e:
        logger.exception(
            f"An error occurred during the EnhancedSyntheticEmbeddingLayer example: {str(e)}")
